package sun.nio.fs;

import java.util.concurrent.ExecutionException;
import sun.misc.Unsafe;

abstract class Cancellable implements Runnable
{
    private static final Unsafe unsafe;
    private final long pollingAddress;
    private final Object lock;
    private boolean completed;
    private Throwable exception;
    
    protected Cancellable() {
        this.lock = new Object();
        this.pollingAddress = Cancellable.unsafe.allocateMemory(4L);
        Cancellable.unsafe.putIntVolatile(null, this.pollingAddress, 0);
    }
    
    protected long addressToPollForCancel() {
        return this.pollingAddress;
    }
    
    protected int cancelValue() {
        return Integer.MAX_VALUE;
    }
    
    final void cancel() {
        synchronized (this.lock) {
            if (!this.completed) {
                Cancellable.unsafe.putIntVolatile(null, this.pollingAddress, this.cancelValue());
            }
        }
    }
    
    private Throwable exception() {
        synchronized (this.lock) {
            return this.exception;
        }
    }
    
    @Override
    public final void run() {
        try {
            this.implRun();
        }
        catch (Throwable exception) {
            synchronized (this.lock) {
                this.exception = exception;
            }
        }
        finally {
            synchronized (this.lock) {
                this.completed = true;
                Cancellable.unsafe.freeMemory(this.pollingAddress);
            }
        }
    }
    
    abstract void implRun() throws Throwable;
    
    static void runInterruptibly(Cancellable task) throws ExecutionException {
        Thread thread = new Thread(task);
        thread.start();
        boolean n = false;
        while (thread.isAlive()) {
            try {
                thread.join();
            }
            catch (InterruptedException ex) {
                n = true;
                task.cancel();
            }
        }
        if (!(n == 0)) {
            Thread.currentThread().interrupt();
        }
        Throwable exception = task.exception();
        if (exception != null) {
            throw new ExecutionException(exception);
        }
        return;
    }
    
    static {
        unsafe = Unsafe.getUnsafe();
    }
}
